/*
 *
 * MODIFIED GADGETFS EXAMPLE TO WORK AS A SIMULATED USB MOUSE FOR THE 
 * PXA255 PROCESSOR
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <memory.h>
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <asm/byteorder.h>

#include <linux/types.h>
#include <linux/usb/gadgetfs.h>
#include <linux/usb/ch9.h>
//#include <linux/hid.h>

#include "usbstring.h"
//#include <usb.h>

static int verbose;

#define	iosize 8

static unsigned long fill_in_buf(void *buf, unsigned long nbytes)
{
	__u8 sbuf[8] = {0x00, 0x00, 0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
	memcpy(buf, sbuf, iosize);
	return iosize;
}

static unsigned long fill_in_buf_zero(void *buf, unsigned long nbytes)
{
	__u8 sbuf[8] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	memcpy(buf, sbuf, iosize);
	return iosize;
}

#define DRIVER_VENDOR_NUM	0x046d		/* Logitech */

#define DRIVER_PRODUCT_NUM	0xc312 


#define	STRINGID_MFGR		0x1
#define	STRINGID_PRODUCT	2
#define	STRINGID_SERIAL		0
#define	STRINGID_CONFIG		0
#define	STRINGID_INTERFACE	0

int int_fd;
static struct usb_device_descriptor device_desc = {
	.bLength =		sizeof device_desc,
	.bDescriptorType =	USB_DT_DEVICE,
	.bDeviceClass =		0,
	.bDeviceSubClass =	0,
	.bDeviceProtocol =	0,
	// .bMaxPacketSize0 ... set by gadgetfs
	.idVendor =		__constant_cpu_to_le16 (DRIVER_VENDOR_NUM),
	.idProduct =		__constant_cpu_to_le16 (DRIVER_PRODUCT_NUM),
	.iManufacturer =	STRINGID_MFGR,
	.iProduct =		STRINGID_PRODUCT,
	.iSerialNumber =	STRINGID_SERIAL,
	.bNumConfigurations =	1,
};

#define	MAX_USB_POWER		1

#define	CONFIG_VALUE		1

static const struct usb_config_descriptor
config = {
	.bLength =		sizeof config,
	.bDescriptorType =	USB_DT_CONFIG,

	/* must compute wTotalLength ... */
	.bNumInterfaces =	1,
	.bConfigurationValue =	CONFIG_VALUE,
	.iConfiguration = 	STRINGID_CONFIG,
	.bmAttributes =		USB_CONFIG_ATT_ONE |
					 USB_CONFIG_ATT_SELFPOWER,
	.bMaxPower =		0x32//(MAX_USB_POWER + 1) / 2,
};

static struct usb_interface_descriptor
source_sink_intf = {
	.bLength =		sizeof source_sink_intf,
	.bDescriptorType =	USB_DT_INTERFACE,

	.bInterfaceClass =	USB_CLASS_HID,
	.bInterfaceSubClass =	1,
	.bInterfaceProtocol =	1,
	.iInterface =		STRINGID_INTERFACE,
};

// /usr/src/kernels/2.6.30.5-43.fc11.x86_64/include/linux/hid.h
struct hid_class_descriptor {
	__u8 bDescriptorType;
	__le16 wDescriptorLength;
} __attribute__ ((packed));

struct hid_descriptor {
	__u8 bLength;
	__u8 bDescriptorType;
	__le16 bcdHID;
	__u8 bCountryCode;
	__u8 bNumDescriptors;

	struct hid_class_descriptor desc[1];
} __attribute__ ((packed));



static struct hid_descriptor
hid_desc = {
	.bLength = 9,
	.bDescriptorType = USB_DT_CS_DEVICE,
	.bcdHID = __constant_cpu_to_le16(0x0110),
	.bCountryCode = 0,
	.bNumDescriptors = 1,
};



/*
const char ReportDescriptor1[] = {
	0x05, 0x01, //Usage Page (Generic Desktop)
	0x09, 0x02, //Usage (Mouse)
	0xA1, 0x01, //Collection (Application)
	0x09, 0x01, //Usage (Pointer)
	0xA1, 0x00, //Collection (Physical)
	0x05, 0x09, //Usage Page (Buttons)
	0x19, 0x01, //Usage Minimum (1)
	0x29, 0x03, //Usage Maximum (3)
	0x15, 0x00, //Logical Minimum (0)
	0x25, 0x01, //Logical Maximum (1)
	0x95, 0x03, //Report Count (3)
	0x75, 0x01, //Report Size (1)
	0x81, 0x02, //Input (Data, Variable, Absolute)
	0x95, 0x01, //Report Count (1)
	0x75, 0x05, //Report Size (5)
	0x81, 0x01, //Input (03=Logitech, 01=Hid spec:Constant)
	0x05, 0x01, //Usage Page (Generic Desktop)
	0x09, 0x30, //Usage (X)
	0x09, 0x31, //Usage (Y)
	0x09, 0x38, //Usage (Logitech:_)
	0x15, 0x81, //Logical Minimum (-127)
	0x25, 0x7F, //Logical Maximum (127)
	0x75, 0x08, //Report Size (8)
	0x95, 0x02, //Report Count (Logitech=3, Hid spec =2)
	0x81, 0x06, //Input (Data, Variable, Relative)
	0xC0, 	//End Collection
	0xC0	// End Collection
}; */

const char ReportDescriptor1[] = "\x05\x01\x09\x06\xa1\x01\x05\x07\x19\xe0\x29\xe7\x15\x00\x25\x01\x75\x01\x95\x08\x81\x02\x95\x01\x75\x08\x81\x01\x95\x03\x75\x01\x05\x08\x19\x01\x29\x03\x91\x02\x95\x05\x75\x01\x91\x01\x95\x06\x75\x08\x15\x00\x26\xff\x00\x05\x07\x19\x00\x2a\xff\x00\x81\x00\xc0"; 

const unsigned int ReportDescSize = sizeof(ReportDescriptor1) - 1;

const char MouseMoveArrayX[] = {
	0x8F, 0x0F,
	0x00, 0x00,
	0x8F, 0x00,
	0x8F, 0x00,
	0x00
};

const unsigned int MouseMoveArraySize = sizeof(MouseMoveArrayX);

const char MouseMoveArrayY[] = {
	0x00, 0x00,
	0x3F, 0x9F,
	0x3F, 0x00,
	0x00, 0x00,
	0x00
};


const char MouseMoveArrayButtons[] = {
	0x00, 0x00,
	0x00, 0x02,
	0x02, 0x00,
	0x00, 0x01,
	0x00
};

/* some devices can handle other status packet sizes */
#define STATUS_MAXPACKET	8
#define	LOG2_STATUS_POLL_MSEC	3

static struct usb_endpoint_descriptor
fs_status_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16 (iosize),
	.bInterval =		0x0A,
};

static const struct usb_endpoint_descriptor *fs_eps [] = {
	&fs_status_desc,
};

static struct usb_endpoint_descriptor
hs_status_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16 (iosize),//__constant_cpu_to_le16 (STATUS_MAXPACKET),
	.bInterval =		5,
};

static const struct usb_endpoint_descriptor *hs_eps [] = {
	&hs_status_desc,
};

/*-------------------------------------------------------------------------*/

static char serial [32];

static struct usb_string stringtab [] = {
	{ STRINGID_MFGR,	"Licensed to Code, LLC", },
	{ STRINGID_PRODUCT,	"My USB Keyboard", },
	{ STRINGID_SERIAL,	serial, },
	{ STRINGID_CONFIG,	"The Configuration", },
	{ STRINGID_INTERFACE,	"Source/Sink", },
};

static struct usb_gadget_strings strings = {
	.language =	0x0409,		/* "en-us" */
	.strings =	stringtab,
};

/*-------------------------------------------------------------------------*/

/* kernel drivers could autoconfigure like this too ... if
 * they were willing to waste the relevant code/data space.
 */

static int	HIGHSPEED;
static char	*DEVNAME;
static char	*EP_STATUS_NAME;

/* gadgetfs currently has no chunking (or O_DIRECT/zerocopy) support
 * to turn big requests into lots of smaller ones; so this is "small".
 */
#define	USB_BUFSIZE	(7 * 1024)

static enum usb_device_speed	current_speed;

static inline int min(unsigned a, unsigned b)
{
	return (a < b) ? a : b;
}

static int autoconfig ()
{
	struct stat	statb;

	/* Intel PXA 2xx processor, full speed only */
	if (stat (DEVNAME = "dummy_udc", &statb) == 0) {
		HIGHSPEED = 1;
		device_desc.bcdDevice = __constant_cpu_to_le16 (0x0100);

		EP_STATUS_NAME = "ep-f";

		source_sink_intf.bNumEndpoints = 1;
		hs_status_desc.bEndpointAddress =
		fs_status_desc.bEndpointAddress = 0x81; //Logitech=82, HId spec=81
	}
	else {
		DEVNAME = 0;
		return -ENODEV;
	}
	return 0;
}


/*-------------------------------------------------------------------------*/

/* full duplex data, with at least three threads: ep0, sink, and source */
/*Only using the ep0 thread for the mouse simulation */

static pthread_t	ep0;

static pthread_t	source;
static int		source_fd = -1;

// FIXME no status i/o yet

static void close_fd (void *fd_ptr)
{
	int	status, fd;

	fd = *(int *)fd_ptr;
	*(int *)fd_ptr = -1;

	/* test the FIFO ioctls (non-ep0 code paths) */
	if (pthread_self () != ep0) {
		status = ioctl (fd, GADGETFS_FIFO_STATUS);
		if (status < 0) {
			/* ENODEV reported after disconnect */
			if (errno != ENODEV && errno != -EOPNOTSUPP)
				perror ("get fifo status");
		} else {
			fprintf (stderr, "fd %d, unclaimed = %d\n",
				fd, status);
			if (status) {
				status = ioctl (fd, GADGETFS_FIFO_FLUSH);
				if (status < 0)
					perror ("fifo flush");
			}
		}
	}

	if (close (fd) < 0)
		perror ("close");
}

/* you should be able to open and configure endpoints
 *  * whether or not the host is connected
 *   */
static int
ep_config (char *name, const char *label,
        struct usb_endpoint_descriptor *fs,
        struct usb_endpoint_descriptor *hs
)
{
        int             fd, status;
        char            buf [USB_BUFSIZE];

        /* open and initialize with endpoint descriptor(s) */
        fd = open (name, O_RDWR);
        if (fd < 0) {
                status = -errno;
                fprintf (stderr, "%s open %s error %d (%s)\n",
                        label, name, errno, strerror (errno));
                return status;
        }

        /* one (fs or ls) or two (fs + hs) sets of config descriptors */
        *(__u32 *)buf = 1;      /* tag for this format */
        memcpy (buf + 4, fs, USB_DT_ENDPOINT_SIZE);
        if (HIGHSPEED)
                memcpy (buf + 4 + USB_DT_ENDPOINT_SIZE,
                        hs, USB_DT_ENDPOINT_SIZE);
        status = write (fd, buf, 4 + USB_DT_ENDPOINT_SIZE
                        + (HIGHSPEED ? USB_DT_ENDPOINT_SIZE : 0));
        if (status < 0) {
                status = -errno;
                fprintf (stderr, "%s config %s error %d (%s)\n",
                        label, name, errno, strerror (errno));
                close (fd);
                return status;
        } else if (verbose) {
                unsigned long   id;

                id = pthread_self ();
                fprintf (stderr, "%s start %ld fd %d\n", label, id, fd);
        }
        return fd;
}

#define status_open(name) \
        ep_config(name,__FUNCTION__, &fs_status_desc, &hs_status_desc)


static void *simple_source_thread (void *param)
{
	char		*name = (char *) param;
	int		status;
	char		buf [USB_BUFSIZE];

	status = status_open (name);
	if (status < 0)
		return 0;

	source_fd = status;

	pthread_cleanup_push (close_fd, &source_fd);
	sleep(5);
	do {
		sleep(5);
		printf("%s sending event\n", name);
		unsigned long	len;

		/* original LinuxThreads cancelation didn't work right
		 * so test for it explicitly.
		 */
		pthread_testcancel ();

		len = fill_in_buf (buf, sizeof buf);
		if (len > 0)
			status = write (source_fd, buf, len);
		else
			status = 0;
		printf("%s sent event %d %m\n", name, status);

		len = fill_in_buf_zero (buf, sizeof buf);
		if (len > 0)
			status = write (source_fd, buf, len);
		else
			status = 0;
		printf("%s sent event %d %m\n", name, status);


	} while (1);//(status > 0);
	if (status == 0) {
		if (verbose)
			fprintf (stderr, "done %s\n", __FUNCTION__);
	} else if (verbose > 2 || errno != ESHUTDOWN) /* normal disconnect */
		perror ("write");
	fflush (stdout);
	fflush (stderr);
	pthread_cleanup_pop (1);

	return 0;
}

/*-------------------------------------------------------------------------*/
static void start_io ()
{
	sigset_t	allsig, oldsig;

	sigfillset (&allsig);
	errno = pthread_sigmask (SIG_SETMASK, &allsig, &oldsig);
	if (errno < 0) {
		perror ("set thread signal mask");
		return;
	}

	/* give the other threads a chance to run before we report
	 * success to the host.
	 * FIXME better yet, use pthread_cond_timedwait() and
	 * synchronize on ep config success.
	 */

        if (pthread_create (&source, 0,
                        simple_source_thread, (void *) EP_STATUS_NAME) != 0) {
                perror ("can't create source thread");
                goto cleanup;
        }


	sched_yield ();

cleanup:
	//printf("start of start_io cleanup\n");
	errno = pthread_sigmask (SIG_SETMASK, &oldsig, 0);
	if (errno != 0) {
		perror ("restore sigmask");
		exit (-1);
	}
	printf("exiting start_io cleanup\n");
}

static void stop_io ()
{
	if (!pthread_equal (source, ep0)) {
		pthread_cancel (source);
		if (pthread_join (source, 0) != 0)
			perror ("can't join source thread");
		source = ep0;
	}
}

/*-------------------------------------------------------------------------*/

static char *
build_config (char *cp, const struct usb_endpoint_descriptor **ep)
{
	struct usb_config_descriptor *c;
	int i;

	c = (struct usb_config_descriptor *) cp;

	memcpy (cp, &config, config.bLength);
	cp += config.bLength;
	memcpy (cp, &source_sink_intf, source_sink_intf.bLength);
	cp += source_sink_intf.bLength;
	memcpy (cp, &hid_desc, hid_desc.bLength);
	cp += hid_desc.bLength;

	for (i = 0; i < source_sink_intf.bNumEndpoints; i++) {
		memcpy (cp, ep [i], USB_DT_ENDPOINT_SIZE);
		cp += USB_DT_ENDPOINT_SIZE;
	}
	c->wTotalLength = __cpu_to_le16 (cp - (char *) c);
	return cp;
}

static int init_device (void)
{
	char		buf [4096], *cp = &buf [0];
	int		fd;
	int		status;

	status = autoconfig ();
	if (status < 0) {
		fprintf (stderr, "?? don't recognize device: %m\n");
		return status;
	}

	fd = open (DEVNAME, O_RDWR);
	if (fd < 0) {
		perror (DEVNAME);
		return -errno;
	}

	*(__u32 *)cp = 0;	/* tag for this format */
	cp += 4;

	/* write full then high speed configs */
	cp = build_config (cp, fs_eps);
	if (HIGHSPEED)
		cp = build_config (cp, hs_eps);

	/* and device descriptor at the end */
	memcpy (cp, &device_desc, sizeof device_desc);
	cp += sizeof device_desc;

	status = write (fd, &buf [0], cp - &buf [0]);
	printf("status for init_device = %x\n", status);
	if (status < 0) {
		perror ("write dev descriptors");
		close (fd);
		return status;
	} else if (status != (cp - buf)) {
		fprintf (stderr, "dev init, wrote %d expected %d\n",
				status, cp - buf);
		close (fd);
		return -EIO;
	}
	return fd;
}

static void handle_control (int fd, struct usb_ctrlrequest *setup, int s_iCounter)
{
	int		status, tmp;
	__u8		buf [256];
	int		i;
	i = 0;

	__u16		value, index, length;

	value = __le16_to_cpu(setup->wValue);
	index = __le16_to_cpu(setup->wIndex);
	length = __le16_to_cpu(setup->wLength);

		fprintf (stderr, "s_iCounter = %x\n SETUP %02x.%02x "
				"v%04x i%04x %d\n",
			s_iCounter, setup->bRequestType, setup->bRequest,
			value, index, length);

	printf("fd in handle_control = %x\n", fd);
	//fd is the same each time this function is called

	switch (setup->bRequest) {	/* usb 2.0 spec ch9 requests */
	case USB_REQ_GET_DESCRIPTOR:
		printf("In handle_control Get Descriptor case\n");

		switch (value >> 8) {
		case USB_DT_STRING:
			printf("get descriptor string type\n");
			if (setup->bRequestType != USB_DIR_IN)
				goto stall;
			tmp = value & 0x0ff;
			if (verbose > 1)
				fprintf (stderr,
					"... get string %d lang %04x\n",
					tmp, index);
			if (tmp != 0 && index != strings.language)
				goto stall;
			status = usb_gadget_get_string (&strings, tmp, buf);
			if (status < 0)
				goto stall;
			tmp = status;
			if (length < tmp)
				tmp = length;
			status = write (fd, buf, tmp);
			if (status < 0) {
				if (errno == EIDRM)
					fprintf (stderr, "string timeout\n");
				else
					perror ("write string data");
			} else if (status != tmp) 
				fprintf (stderr, "short string write, %d\n",
				status);
			break;
		case 0x22:
			printf("Got an report descriptor request\n");
			tmp = value & 0x00ff;
			if (verbose > 1)
				printf("... get rpt desc %d, index %04x\n", tmp, setup->wIndex);
			if (tmp != 0 && setup->wIndex != 0) 
				goto stall;

			memcpy(buf, ReportDescriptor1, ReportDescSize);
			status=write(fd, buf, ReportDescSize);
			printf("status for write report descriptor = %x\n", status);
			if (status < 0)
			{
				if (errno == EIDRM)
					printf ("GET_REPORT timeout\n");
				else
					perror ("write GET_REPORT data");
			}
			break;
		default:
			goto stall;
		}
		return;

	case USB_REQ_SET_CONFIGURATION:
		printf("Set Configuration case\n");
		if (setup->bRequestType != USB_DIR_OUT)
			goto stall;
		if (verbose)
			fprintf (stderr, "CONFIG #%d\n", value);

		/* Kernel is normally waiting for us to finish reconfiguring
		 * the device.
		 *
		 * Some hardware can't, notably older PXA2xx hardware.(With
		 * racey and restrictive config change automagic.)To handle
		 * such hardware, don't write code this way ... instead, keep
		 * the endpoints always active and don't rely on seeing any
		 * config change events, either this or SET_INTERFACE.
		 */
		switch (value) {
		case CONFIG_VALUE:
			printf("about to call start_io function\n");
			start_io ();
			break;
		case 0:
			//printf("about to call stop_io function\n");
			stop_io ();
			//printf("just finished stop_io function\n");
			break;
		default:
			/* kernel bug -- "can't happen" */
			fprintf (stderr, "? illegal config\n");
			goto stall;
		}

		/* ... ack (a write would stall) */
		status = read (fd, &status, 0);
		printf("status of read = %x\n", status);
		if (status)
			perror ("ack SET_CONFIGURATION");
		//printf("About to return from set configuration\n");
		return;
	case USB_REQ_GET_INTERFACE:
		printf("Get Interface case\n");
		if (setup->bRequestType != (USB_DIR_IN|USB_RECIP_INTERFACE)
				|| index != 0
				|| length > 1)
			goto stall;

		/* only one altsetting in this driver */
		buf [0] = 0;
		status = write (fd, buf, length);
		if (status < 0) {
			if (errno == EIDRM)
				fprintf (stderr, "GET_INTERFACE timeout\n");
			else
				perror ("write GET_INTERFACE data");
		} else if (status != length) {
			fprintf (stderr, "short GET_INTERFACE write, %d\n",
				status);
		}
		return;
	case USB_REQ_SET_INTERFACE:
		printf("Set Interface case\n");
		if (setup->bRequestType != USB_RECIP_INTERFACE
				|| index != 0
				|| value != 0)
			goto stall;

		/* just reset toggle/halt for the interface's endpoints */
		status = 0;
		if (ioctl (source_fd, GADGETFS_CLEAR_HALT) < 0) {
			status = errno;
			perror ("reset source fd");
		}
		/* FIXME eventually reset the status endpoint too */
		if (status)
			goto stall;

		/* ... and ack (a write would stall) */
		status = read (fd, &status, 0);
		if (status)
			perror ("ack SET_INTERFACE");
		return;
	default:
		goto stall;
	}

stall:
	if (verbose)
		printf ("... protocol stall %02x.%02x\n",
			setup->bRequestType, setup->bRequest);

	/* non-iso endpoints are stalled by issuing an i/o request
	 * in the "wrong" direction.ep0 is special only because
	 * the direction isn't fixed.
	 */
	if (setup->bRequestType & USB_DIR_IN)
		status = read (fd, &status, 0);
	else
		status = write (fd, &status, 0);
	if (status != -1)
		printf ("can't stall ep0 for %02x.%02x\n",
			setup->bRequestType, setup->bRequest);
	else if (errno != EL2HLT)
		perror ("ep0 stall");

}


static void signothing (int sig, siginfo_t *info, void *ptr)
{
	/* NOP */
	if (verbose > 2)
		fprintf (stderr, "%s %d\n", __FUNCTION__, sig);
}

static const char *speed (enum usb_device_speed s)
{
	switch (s) {
	case USB_SPEED_LOW:	return "low speed";
	case USB_SPEED_FULL:	return "full speed";
	case USB_SPEED_HIGH:	return "high speed";
	default: 		return "UNKNOWN speed";
	}
}

/*-------------------------------------------------------------------------*/

/* control thread, handles main event loop*/

#define	NEVENT	5

static void *ep0_thread (void *param)
{
	int			fd = *(int*) param;
	struct sigaction	action, old;
	sigset_t		sigs;
	static int		s_iCounter = 0;

	source = ep0 = pthread_self ();
	pthread_cleanup_push (close_fd, param);

	/* SIGIO tells file owner that a device event is available
	 * use F_SETSIG to use a different signal.
	 * (not SIGUSR1 or SIGUSR2; linux pthreads may use them.)
	 * right now, only ep0 supports O_ASYNC signaling.
	 */
	if (fcntl (fd, F_SETOWN, getpid ()) < 0) {
		perror ("set ep0 owner");
		return 0;
	}
	if (fcntl (fd, F_SETFL, O_ASYNC) < 0) {
		perror ("set ep0 async mode");
		return 0;
	}
	action.sa_sigaction = signothing;
	sigfillset (&action.sa_mask);
	action.sa_flags = SA_SIGINFO;
	if (sigaction (SIGIO, &action, &old) < 0) {
		perror ("sigaction");
		return 0;
	}
	/* event loop */
	sigfillset (&sigs);
	for (;;) {
		int				tmp;
		struct usb_gadgetfs_event	event [NEVENT];
		int				connected = 0;
		int				i, nevent;

		sigwait (&sigs, &tmp);

		/* eventually any endpoint could send signals */
		switch (tmp) {
		case SIGIO:		/* used only by ep0 */
			printf("EP0 thread SIGIO case\n");
			tmp = read (fd, &event, sizeof event);
			if (tmp < 0) {
				printf("tmp < 0\n");
				if (errno == EAGAIN) {
					sleep (1);
					continue;
				}
				perror ("ep0 read after sigio");
				goto done;
			}
			nevent = tmp / sizeof event [0];
			if (verbose)
				fprintf (stderr, "read %d ep0 events\n",
					nevent);

			for (i = 0; i < nevent; i++) {
				switch (event [i].type) {
				case GADGETFS_NOP:
					printf("NOP\n");
					if (verbose)
						fprintf (stderr, "NOP\n");
					break;
				case GADGETFS_CONNECT:
					connected = 1;
					printf("CONNECTED\n");
					current_speed = event [i].u.speed;
					if (verbose)
						fprintf (stderr,
							"CONNECT %s\n",
						speed (event [i].u.speed));
					break;
				case GADGETFS_SETUP:
					s_iCounter++;
					connected = 1;
					handle_control (fd, &event [i].u.setup, s_iCounter);
					break;
				case GADGETFS_DISCONNECT:
					connected = 0;
					printf("DISCONNECTED\n");
					current_speed = USB_SPEED_UNKNOWN;
					if (verbose)
						fprintf(stderr, "DISCONNECT\n");
					stop_io ();
					break;
				case GADGETFS_SUSPEND:
					printf("SUSPEND\n");
					if (verbose)
						fprintf (stderr, "SUSPEND\n");
					break;
				default:
					printf("default case\n");
					fprintf (stderr,
						"* unhandled event %d\n",
						event [i].type);
				}
			}
			continue;

		case SIGINT:
			break;
		case SIGTERM:
			fprintf (stderr, "\nquit, sig %d\n", tmp);
			break;
		case SIGUSR1:
			break;
		case SIGUSR2:
			fprintf (stderr, "\npthreads sig %d\n", tmp);
			break;
		default:
			fprintf (stderr, "\nunrecognized signal %d\n", tmp);
			break;
		}
done:
		fflush (stdout);
		if (connected)
		{
			stop_io ();
		}
		break;
	}

	if (verbose)
		fprintf (stderr, "done\n");
	fflush (stdout);

	pthread_cleanup_pop (1);
	return 0;
}

/*-------------------------------------------------------------------------*/
int main (int argc, char **argv)
{
	int		fd, c, i;

	hid_desc.desc->bDescriptorType = 0x22;
	hid_desc.desc->wDescriptorLength = 65;


	/* random initial serial number */
	srand ((int) time (0));
	for (i = 0; i < sizeof serial - 1; ) {
		c = rand () % 127;
		if ((('a' <= c && c <= 'z') || ('0' <= c && c <= '9')))
			serial [i++] = c;
	}

	while ((c = getopt (argc, argv, "I:a:i:o:r:s:v")) != EOF) {
		switch (c) {
		case 'r':		/* nonrandom serial */
			strncpy (serial, optarg, sizeof serial - 1);
			continue;
		case 'v':		/* verbose */
			verbose++;
			continue;
		}
		fprintf (stderr, "usage:%s "
				"[-r serial] [-v]\n",
				argv [0]);
		return 1;
	}
	if (chdir ("/dev/gadget") < 0) {
		perror ("can't chdir /dev/gadget");
		return 1;
	}

	fd = init_device ();
	if (fd < 0)
		return 1;

	fprintf (stderr, "/dev/gadget/%s ep0 configured\nserial=\"%s\"\n",
		DEVNAME, serial);
	fflush (stderr);
	(void) ep0_thread (&fd);
	return 0;
}
